/*
 * Copyright (c) 2010, Noel Lemouel
 * All rights reserved.
 */

#include <sysutil.h>
#include <string.h>

#define ETX 0x03
#define BS  0x08
#define TAB 0x09
#define LF  0x0a
#define CR  0x0d
#define ESC 0x1b
#define DEL 0x7f

#define MAX_SH_LINE_SIZE 256
#define MAX_SH_ARG_SIZE 16

#define MAX_SH_P_ARG (MAX_SH_LINE_SIZE/MAX_SH_ARG_SIZE)

static unsigned int x = 0;  /* exit */

/* internal commands [forwards] */

int __exit(int argc, char **argv);
int __echo(int argc, char **argv);

/* external commands */

extern int _peek(int argc, char **argv);
extern int _poke(int argc, char **argv);
extern int _tp(int argc, char **argv);

/* ... */

struct cmd_s {
   const char *name;
   int (*main)(int argc, char **argv);
};

static const struct cmd_s cmd[] = {
/* builtin */
   {"echo", __echo},
   {"exit", __exit},
/* external */
   {"peek", _peek},
   {"poke", _poke},
   {"tp", _tp},
   {0}
};

/* builtin commands */

int __exit(int argc, char **argv)
{
   x = 1;
   return 0;
}

int __echo(int argc, char **argv)
{
   int i=2;

   if(argc < 2) return -1;

   printf("%s", argv[1]);
   argc-=2;

   while(argc--)
   {
      printf(" ");
      printf("%s", argv[i]);
      i++;
   }

   printf("\n");

   return 0;
}

/* ... */

static int get_line(unsigned char *line)
{
   unsigned int i;
   unsigned char c;

   i = 0;
   while(i<MAX_SH_LINE_SIZE)
   {
      i++;
      c = getchar();
// printf("0x%02x\n", c);
// printf("%d\n", i);
      switch(c){
         case BS:
         case DEL:
            if(i>1){
               i-=2;
               putchar(BS);
               putchar(' ');
               putchar(BS);
            }else{
               i=0;
            }
            break;
         case TAB:
            line[i-1] = ' ';
            putchar(' ');
            break;
         case CR:
         case LF:
            line[i-1] = c;
            return i;
            break;
         case ETX:
            return 0;
            break;
         default:
            line[i-1] = c;
            putchar(c);
            break;
      }
   }

   return i;
}

static int parse(unsigned char *line, unsigned char **argv)
{
   unsigned int argc;
   unsigned int i, j, k;
   unsigned char c;

   i=0;
   j=0;
   k=0;
   c=0;

   argc = 0;

// printf("\n");
   while(i<MAX_SH_LINE_SIZE)
   {
// printf("-   %d %d %d -%c- 0x%02x\n", i, j, k, line[i], line[i]);
      if((line[i] == ' ') || (k == MAX_SH_ARG_SIZE))
      {
         if(k){
// printf("--  %d %d %d\n", i, j, k);
            argv[j][k] = '\0';
            i++; j++; k=0; argc++;
            if(j == MAX_SH_P_ARG){
               return argc;
            }
         }
         while(line[i] == ' ') i++;
      }
      if((line[i] == '#') || (line[i] == LF) || (line[i] == CR))
      {
         argv[j][k] = '\0';
         if(k){
            argc++;
         }
         break;
      }
      argv[j][k] = line[i];
      i++; k++;
   }

   return argc;
}

static void parse_debug(int argc, char *argv[])
{
   int i;

   printf("\n");
   printf("argc: %d\n", argc);

   for(i=0; i<argc; i++)
   {
      printf("argv[%d] = %s-\n", i, argv[i]);
   }
}

static int exec(int argc, unsigned char *argv[])
{
   int i=0;

   if(!argc) return -1;

   printf("\n");

   while(cmd[i].name)
   {
      if(!strcmp((const char*)argv[0], (const char*)cmd[i].name))
      {
         cmd[i].main(argc, (char**)argv);
         return 0;
      }
      // printf("%s\n", cmd[i].name);
      i++;
   }
   printf("ush: %s: command not found\n", argv[0]);

   return 0;
}

int ush()
{
   unsigned int i;
   unsigned char line[MAX_SH_LINE_SIZE];

   unsigned char* argv_p_mem[MAX_SH_P_ARG];
   unsigned char  argv_data_mem[MAX_SH_P_ARG * MAX_SH_ARG_SIZE];

   int argc = 0;
   unsigned char **argv;

   argv = argv_p_mem;
   argv[0] = argv_data_mem;
   for(i=1; i<MAX_SH_P_ARG; i++)
   {
      argv[i] = argv[0] + i * MAX_SH_ARG_SIZE;
   }

// printf("ush\n");
   printf("\n");

   while(!x)
   {
      printf("$ ");
      i = get_line(line);
      if(i) argc = parse(line, argv);
   // parse_debug(argc, (char**)argv);
      if(argc && i){
         exec(argc, argv);
      }else{
         printf("\n");
      }
   }

   return 0;
}

